#if (MACHINE == X68000)
/*
 * MOUSE DRIVER for Minix/X68k.
 *
 * This driver accepts following messages:
 *
 *    m_type     DEVICE    PROC_NR    COUNT    ADRRESS   TTY_FLAGS
 * ----------------------------------------------------------------
 * | HARD_INT  |         |         |         |         |          |
 * ----------------------------------------------------------------
 * | TTY_READ  | device  | proc nr |  bytes  | buf ptr | nonblock |
 * ----------------------------------------------------------------
 * | TTY_IOCTL | device  | proc nr | func nr | buf ptr |          |
 * ----------------------------------------------------------------
 * | CANCEL    | device  | proc nr |         |         |          |
 * ----------------------------------------------------------------
 *
 *
 * This file contains following entry points:
 *
 *	mouse_task:	main entry when system is brought up
 *	msint:		mouse interrupt handler (Rx data)
 *	mserr:		mouse interrupt handler (Rx error)
 *	mstimer:	called by clock task
 *
 */

#include "kernel.h"
#include <minix/callnr.h>
#include <minix/com.h>
#include <minix/mouse.h>
#include "proc.h"
#include "x6addr.h"
#include "x6mfp.h"
#include "x6scc.h"

#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif


/* configurations */
#define DEF_XMAX 30000	/* default value for msctrl.xmax */
#define DEF_YMAX 30000	/* default value for msctrl.ymax */
#define M_INTERVAL 2	/* mouse data is read HZ/(M_INTERVAL+1) times/sec */


/* X68000 mouse packet (2 buttons); 3 bytes */
struct mspacket {
  unsigned char stat;
  char dx, dy;
};
#define M_X6_PACKET_SZ	3	/* mouse packet size (in bytes) */
#define M_X6_BUT_R	0x01	/* button R pressed */
#define M_X6_BUT_L	0x02	/* button L pressed */
#define M_X6_OVFL_DX	0x10	/* DX over flow */
#define M_X6_UDFL_DX	0x20	/* DX under flow */
#define M_X6_OVFL_DY	0x40	/* DY over flow */
#define M_X6_UDFL_DY	0x80	/* DY under flow */


/* local variables */
PRIVATE message mess;			/* message buffer */
PRIVATE struct mspacket mspck;		/* raw mouse packet */
PRIVATE char *mspckp = (char*)&mspck;
PRIVATE int msleft = M_X6_PACKET_SZ;	/* bytes to be received */
PRIVATE int msreq_enabled;		/* MSCTRL line enabled? */

PRIVATE int msready;			/* mouse data available? */
PRIVATE struct mouse msdata;		/* mouse data */
PRIVATE struct mouse oldmsdata;		/* mouse data of previous time step */
PRIVATE phys_bytes msaddr;		/* physical address of msdata */
PRIVATE struct mouse_ctrl msctrl; 	/* mouse ioctrl info */

PRIVATE int mswaiting;			/* someone is waiting the mouse data? */
PRIVATE int mscaller;			/* who made the call (usually FS) */
PRIVATE phys_bytes usr_phys;		/* phys. addr. where mouse data is put */


FORWARD void mouse_init();
FORWARD int do_read();
FORWARD int do_ioctl();
FORWARD void do_intr();
FORWARD void msreq_enable();
FORWARD void msreq_disable();


/* SCC channel B initializing data */
#define DEF_WR3 0xc1
PRIVATE unsigned char SCC_INIT_DAT[] = {
	/***** MODES *****/
	0x01,0x00,		/*** interrupt disable ***/
	0x03,(DEF_WR3 & 0xfe),	/*** Rx 8bits/char, line disable ***/
	0x0b,0x50,		/*** clock select ***/
	0x0c,0x1f,0x0d,0x00,	/*** baud rate 4800 ***/
	0x0e,0x02,0x0e,0x02,	/*** baud rate generator select ***/
	/***** ENABLES *****/
	0x03,DEF_WR3,		/*** Rx 8bits/char, line enable ***/
	0x0e,0x03,		/*** baud rate generator start ***/
	/***** INTERRUPT *****/
	0x0f,0x00,		/*** extend error disable ***/
	0x00,0x10,0x00,0x10,	/*** reset Ext/STATUS ***/
	0x01,0x10,		/*** int on all chars and special condition */
	0xff,0xff,		/*** (end of table) ***/
};

/*===========================================================================*
 *				mouse_task				     *
 *===========================================================================*/
PUBLIC void mouse_task()
{
  int r;
  int procno;

  mouse_init(1);

  while (TRUE) {
      receive(ANY, &mess);

      switch (mess.m_type)
      {
	  case HARD_INT:	do_intr(); break;
	  case TTY_READ:	r = do_read(&mess);	break;
	  case TTY_IOCTL:	r = do_ioctl(&mess);	break;
	  case CANCEL:		r = do_cancel(&mess);	break;
	  default:		r = EINVAL;		break;
      }

      if (mess.m_source > 0) {
	  mess.m_type = TASK_REPLY;
	  mess.REP_PROC_NR = mess.PROC_NR;
	  mess.REP_STATUS = r;
	  send(mess.m_source, &mess);
      }
  }
}

/*===========================================================================*
 *				mouse_init				     *
 *===========================================================================*/
PRIVATE void mouse_init(mode)
int mode;	/* 1:initialize driver  0:set msctrl */
{
  unsigned char *p;
  int s;

  s = lock();
  if (mode == 1)
  {
      msreq_disable();
      msaddr = umap(proc_addr(MOUSE), D, (vir_bytes)&msdata,
      				(vir_bytes)sizeof(struct mouse));
      msctrl.xpos = msctrl.ypos = 0; /* zero the position */
      msctrl.xmickey = 1;	/* default X mickey/pixel ratio */
      msctrl.ymickey = 1;	/* default Y mickey/pixel ratio */
      msctrl.xmax = DEF_XMAX;
      msctrl.ymax = DEF_YMAX;
      msctrl.cursor = FALSE;	/* no cursor updating */
  }

  msdata.xpos = msctrl.xpos;	/* clear the entire data struct */
  msdata.ypos = msctrl.ypos;
  msdata.dx = msdata.dy = 0;
  msdata.buttons = 0;

  mspckp = (char*)&mspck;	/* reset low level packet pointer */
  msleft = M_X6_PACKET_SZ;
  msready = FALSE;		/* mouse data is not ready to read */
  mswaiting = 0;		/* no one waiting for mouse data */

  if (mode == 1)
  {
      /* initialize SCC */
      for (p = SCC_INIT_DAT; *p != 0xff; )
      {
	  SCC->scc_rb = *p++;
	  SCC->scc_rb = *p++;
      }
  }
  restore(s);
}

/*===========================================================================*
 *				do_intr					     *
 *===========================================================================*/
PRIVATE void do_intr()
{
  int s;

  if (mswaiting == 0)
  {
      msready = TRUE;
  }
  else
  {
      /* wake up hanging reader */
      mess.m_type = REVIVE;
      mess.REP_PROC_NR = mswaiting;
      mess.REP_STATUS = sizeof(struct mouse);
      mswaiting = 0;
      msready = FALSE;
      s = lock();
      phys_copy(msaddr, usr_phys, (phys_bytes)sizeof(struct mouse));
      msdata.dx = msdata.dy = 0;
      restore(s);
      send(mscaller, &mess);
  }
}

/*===========================================================================*
 *				do_read					     *
 *===========================================================================*/
PRIVATE int do_read(mp)
message *mp;
{
  int s;
  int r;

  if (mswaiting != 0) /* only 1 process can read mouse */
      return EAGAIN;

  if (mp->COUNT < sizeof(struct mouse))
      return EINVAL;

  usr_phys = umap(proc_addr(mp->PROC_NR), D, (vir_bytes)mp->ADDRESS,
  					(vir_bytes)sizeof(struct mouse));
  if (usr_phys == 0)
      return EFAULT;

  s = lock();

  if (msready) /* mouse data available */
  {
      phys_copy(msaddr, usr_phys, (phys_bytes)sizeof(struct mouse));
      msready = FALSE;
      msdata.dx = msdata.dy = 0;
      r = sizeof(struct mouse);
  }
  else
  {
      if (mp->TTY_FLAGS) /* non blocking */
	  r = 0;
      else
      {
	  mscaller = mp->m_source;
	  mswaiting = mp->PROC_NR;
	  r = SUSPEND;
      }
  }

  restore(s);
  return r;
}

/*===========================================================================*
 *				do_ioctl				     *
 *===========================================================================*/
PRIVATE int do_ioctl(mp)
message *mp;
{
  struct mouse_ctrl tempctrl;
  phys_bytes uphys, mphys;

  switch (mp->COUNT)
  {
  case MSIOCINIT:
      mouse_init(1);
      break;

  case MSIOCGCTRL:
      uphys = umap(proc_addr(mp->PROC_NR), D, (vir_bytes)mp->ADDRESS,
				(vir_bytes)sizeof(struct mouse_ctrl));
      if (uphys == 0)
	return EFAULT;
      mphys = umap(proc_addr(MOUSE), D, (vir_bytes)&msctrl,
  				(vir_bytes)sizeof(struct mouse_ctrl));
      phys_copy(mphys, uphys, (phys_bytes)sizeof(struct mouse_ctrl));
      break;

  case MSIOCSCTRL:
      uphys = umap(proc_addr(mp->PROC_NR), D, (vir_bytes)mp->ADDRESS,
				(vir_bytes)sizeof(struct mouse_ctrl));
      if (uphys == 0)
	return EFAULT;
      mphys = umap(proc_addr(MOUSE), D, (vir_bytes)&tempctrl,
  				(vir_bytes)sizeof(struct mouse_ctrl));
      phys_copy(uphys, mphys, (phys_bytes)sizeof(struct mouse_ctrl));
      if (tempctrl.xpos < 0 || tempctrl.ypos < 0
	  || tempctrl.xmax <= 0 || tempctrl.ymax <= 0
	  || tempctrl.xmickey < 0 || tempctrl.ymickey < 0
	  || tempctrl.xpos > tempctrl.xmax || tempctrl.ypos > tempctrl.ymax
	  || tempctrl.xmickey > tempctrl.xmax || tempctrl.ymickey > tempctrl.ymax)
      {
	return EINVAL;
      }
      msctrl = tempctrl;
      mouse_init(0);
      break;

  default:
      return EINVAL;
      break;
  }
  return OK;
}

/*===========================================================================*
 *				do_cancel				     *
 *===========================================================================*/
/* cancels previous request. */
PRIVATE int do_cancel(mp)
message *mp;
{
  mswaiting = 0;
  return EINTR;
}

/*===========================================================================*
 *				mserr   				     *
 *===========================================================================*/
PUBLIC void mserr(type)
int type;	/* interrupt type */
{
  int s;
  int sink;

  s = lock();

  sink = SCC->scc_db; /* error data */
  SCC->scc_rb = 0x30; /* error reset */
  SCC->scc_rb = 0x38; /* reset highest IUS(intr. under service) bit */

  restore(s);
}

/*===========================================================================*
 *				msint   				     *
 *===========================================================================*/
/* receives mouse data. */
PUBLIC void msint(type)
int type;	/* interrupt type */
{
  short dx, dy;
  int s;

  s = lock();

  while (SCC->scc_rb & 0x01) /* Rx data available */
  {
      *mspckp++ = SCC->scc_db; /* pop a received byte */
      if (--msleft == 0)
	  break;
  }

  if (msleft == 0)
  {
      /* whole of mouse packet was received.
       * Make up mouse struct data, and wake up someone.
       */
      /* make up mouse struct data */
      oldmsdata = msdata;
      dx = (mspck.stat & M_X6_OVFL_DX) ? 128
		    : (mspck.stat & M_X6_UDFL_DX) ? -129
		    : mspck.dx;
      dy = (mspck.stat & M_X6_OVFL_DY) ? 128
		    : (mspck.stat & M_X6_UDFL_DY) ? -129
		    : mspck.dy;
      msdata.dx += dx;
      msdata.dy += dy;
      msdata.xpos += dx;
      if (msdata.xpos < 0)
	  msdata.xpos = 0;
      else if (msdata.xpos > msctrl.xmax)
	  msdata.xpos = msctrl.xmax;
      msdata.ypos += dy;
      if (msdata.ypos < 0)
	  msdata.ypos = 0;
      else if (msdata.ypos > msctrl.ymax)
	  msdata.ypos = msctrl.ymax;
      msdata.buttons =
	    ((mspck.stat & M_X6_BUT_R) ? M_BUT_R : 0)
	  | ((mspck.stat & M_X6_BUT_L) ? M_BUT_L : 0);

      /* wake up someone */
      if (dx < 0) dx = -dx;
      if (dy < 0) dy = -dy;
      if (dx >= msctrl.xmickey || dy >= msctrl.ymickey
	  || msdata.buttons != oldmsdata.buttons)
      {
	  interrupt(MOUSE);
      }
  }

  SCC->scc_rb = 0x38; /* reset highest IUS(intr. under service) bit */
  restore(s);
}

/*===========================================================================*
 *				mstimer 				     *
 *===========================================================================*/
/* requests mouse device for a data packet. */
PUBLIC void mstimer(type)
int type; /* interrupt type */
{
  static int count = M_INTERVAL;
  int s;

  s = lock();

  if (msreq_enabled)
  {
      msreq_disable();
  }
  else if (--count == 0)
  {
      count = M_INTERVAL;
      mspckp = (char*)&mspck;
      msleft = M_X6_PACKET_SZ;
      msreq_enable();
  }

  restore(s);
}

/*===========================================================================*
 *				msreq_enable  				     *
 *===========================================================================*/
/* enables MSCTRL line to request mouse for a packet.
 * The enabling edge of MSCTRL let mouse send a mouse packet.
 */
PRIVATE void msreq_enable()
{
  unsigned char scc_r5b;

  /* already lock()ed */
  msreq_enabled = TRUE;

  while (!((char)(MFP->mf_tsr) & (char)(T_EMPTY)))
      ;
  MFP->mf_udr = 0x40;		/* -MSCTRL(on keyboard):low */

  SCC->scc_rb = 5; scc_r5b = SCC->scc_rb;
  scc_r5b &= ~0x02;		/* -MSCTRL(on body):low */
  SCC->scc_rb = 5; SCC->scc_rb = scc_r5b;

}

/*===========================================================================*
 *				msreq_disable 				     *
 *===========================================================================*/
/* disables MSCTRL line. */
PRIVATE void msreq_disable()
{
  unsigned char scc_r5b;

  /* already lock()ed */
  msreq_enabled = FALSE;

  while (!((char)(MFP->mf_tsr) & (char)(T_EMPTY)))
      ;
  MFP->mf_udr = 0x41;		/* -MSCTRL(on keyboard):high */

  SCC->scc_rb = 5; scc_r5b = SCC->scc_rb;
  scc_r5b |= 0x02;		/* -MSCTRL(on body):high */
  SCC->scc_rb = 5; SCC->scc_rb = scc_r5b;
}

#endif /* MACHINE == X68000 */
